<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief Forms writing and links writing helpers:
/// - #FormElement: form row
/// - #FormManager: form builder and reader
/// - #FormObject: form manager #ObjectBase
//-----------------------------------------------------------------


//-----------------------------------------------------------------
/// Row in a <tt>\<form\></tt> tag.
//-----------------------------------------------------------------

class FormElement
{
	public $sTitle, $sContents, $sComments;
	
	
	public function __construct($sTitle, $sContents, $sComments = '')
	{
		$this->sTitle    = $sTitle;
		$this->sContents = $sContents;
		$this->sComments = $sComments;
	}
	
	public function getRow()
	{
		return
			html::td($this->sTitle, attr::className('form-label')) .
			html::td($this->sContents, attr::className('form-field')) .
			html::td($this->sComments, attr::className('form-comments'));
	}
}


//-----------------------------------------------------------------
/// List of #FormElement objects with helper methods
//-----------------------------------------------------------------

class FormManager
{
	private $aElem = array();
	protected $page;
	
	const SEPARATOR = 'separator';
	
	
	public function __construct(Page $page, array $aElem = array())
	{
		$this->page  = $page;
		$this->aElem = $aElem;
	}
	
	public function html()
	{
		$s = '';
		$bSeparator = false;
		foreach ($this->aElem as $elem)
			if ($elem === self::SEPARATOR)
				$bSeparator = true;
			else if ($elem instanceof FormElement) {
				$s .= html::tr($elem->getRow(), ($bSeparator) ? attr::className('sep-before') : array());
				$bSeparator = false;
			}else
				debug::fatalError('FormManager components must be separators or FormElement objects');
		
		return html::table($s, '', attr::className('form-table'));
	}
	
	
	//-----------------------------------------------------------------
	/// @name Accessors
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Return the current value of a field.
	/// If no value is specified by the HTML request,
	/// return an empty value.
	public function getValue($sName)
	{
		return $_REQUEST->getSafe($sName);
	}
	
	/// @brief Set the value of a field.
	public function setValue($sName, $value)
	{
		$_REQUEST->$sName = $value;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Fields pushing
	/// @{
	//-----------------------------------------------------------------
	
	public function push($elem)
	{
		$this->aElem[] = $elem;
	}
	
	public function pushElement($sLabel, $sContents, $sComments = '')
	{
		$this->push(new FormElement($sLabel, $sContents, $sComments));
	}
	
	public function pushElementLabel($sId, $sLabel, $sContents, $sComments = '')
	{
		$this->pushElement(form::label($sId, $sLabel), $sContents, $sComments);
	}
	
	public function pushMany(array $aElem)
	{
		foreach ($aElem as $elem)
			$this->push($elem);
	}
	
	public function pushSeparator()
	{
		$this->push(self::SEPARATOR);
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Generic fields adding
	///
	/// Each method validates its input and raises an error if necessary.
	/// Default values are taken in the edited object.
	/// @{
	//-----------------------------------------------------------------
	
	/// Add a single-line input field to the form.
	/// @param $sName the name of the field.
	/// @param $maxLength the maximum possible length of the field value, in characters.
	/// An error is raised if the given value is too long.
	/// @param $iSize size of the text field, in em unit.
	/// @param $sLabel the field label, put in the left column.
	/// @param $sComments the field comments, put in the rightmost column.
	/// @param $sDefaultText used to give a default value to the fields
	/// without saving the default value in the database
	/// if the user does not modify it.
	/// Example: <tt>http://</tt> for URLs.
	/// @param $sClassName the HTML class of the \c INPUT anchor.
	/// @param $sMore the HTML code to put after the \c INPUT anchor.
	public function pushText($sName, $maxLength, $iSize, $sLabel, $sComments = '', $sDefaultText = '', $sClassName = 'text', $sMore = '')
	{
		$sValue = trim($this->getValue($sName));
		if ($sValue === $sDefaultText)
			$this->setValue($sName, '');
		
		if (!empty($this->page->bSubmit) && strlen($sValue) > $maxLength)
			$this->error($sLabel . '&nbsp;: texte trop long');
		
		if (isEmpty($sValue))
			$sValue = $sDefaultText;
		
		if ($iSize >= 0 && $sClassName == 'text')
			$sClassName = '';
		$this->pushElementLabel($sName, $sLabel,
			form::text('ID' . $sName, $sValue, $iSize, $maxLength, attr::className($sClassName)) . $sMore, $sComments);
	}
	
	/// Add a multiple-line input field to the form.
	public function pushTextArea($sName, $sLabel, $nbRow, $sComments = '')
	{
		$sValue = $this->getValue($sName);
		$this->pushElementLabel($sName, $sLabel,
			form::textArea('ID' . $sName, -1, $nbRow, html::text($sValue)), $sComments);
	}
	
	
	/// Add a checkbox to the form.
	public function pushCheckbox($sName, $sLabel, $sComments = '', $sClassName = 'checkbox', $sMore = '')
	{
		$bValue = $this->getValue($sName);
		
		$this->pushElementLabel($sName, $sLabel,
			form::checkbox($sName, $bValue, attr::className($sClassName)) . $sMore, $sComments);
	}
	
	/// @}
}


//-----------------------------------------------------------------
/// @brief Enhancement of #FormManager used to build
/// the <tt>\<form\></tt> tag of an object-related form (node, property).
/// Allows to read form values in a safe way.
//-----------------------------------------------------------------

class FormObject extends FormManager
{
	protected $obj;
	protected $aError = array();
	
	protected $aProperty = array();
	protected $aField = array();   // Associations name => value
	
	
	public function __construct(PageDenif $page, ObjectBase $obj)
	{
		parent::__construct($page);
		$this->obj = $obj;
	}
	
	
	protected function loadProperties($iProperty)
	{
		if (!array_key_exists($iProperty, $this->aProperty))
			$this->aProperty[$iProperty] = getProperties($iProperty);
		return $this->aProperty[$iProperty];
	}
	
	
	//-----------------------------------------------------------------
	/// @name Errors handling
	/// @{
	//-----------------------------------------------------------------

	public function error($sError)
	{
		$this->aError[] = $sError;
	}
	
	public function isError() { return !empty($this->aError); }
	
	public function getErrors()
	{
		return html::div(html::mapConcat('div', $this->aError),
			attr::className('errors'));
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Accessors
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Return the array of all fields supported by the form.
	/// Only the fields added by a \c push* method are returned.
	public function getFields() { return $this->aField; }
	
	
	/// @brief Return the current value of a field.
	/// If no value is specified by the HTML request,
	/// return the current edited object value.
	public function getValue($sName)
	{
		$_REQUEST->setDefault($sName, $this->obj->$sName);
		$value = $_REQUEST->getSafe($sName);
		$this->aField[$sName] = $value;
		return $value;
	}
	
	/// @brief Set the value of a field.
	public function setValue($sName, $value)
	{
		parent::setValue($sName, $value);
		$this->aField[$sName] = $value;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Specific fields adding: properties handling
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Add combo-boxes to the form to modify the links of the current node
	/// having the given type (\p $iProperty).
	/// Each combo-box contains all properties of the given type.
	/// HTML links removing the links are provided in the comments column.
	///
	/// Create <tt>aid..._set[]</tt> fields having strictly positive keys.
	/// Each key is the ID of the link to edit.
	///
	/// @param $iProperty the type of the property.
	/// @return the number of properties displayed.
	public function pushEditLinks($iProperty)
	{
		$info = ObjectTypeInfo::get($iProperty);
		$aChoice = $this->getPropertyChoices(false, $iProperty);
		
		$page = $this->page;
		$iIndex = 0;
		foreach ($this->obj->getPropertiesLink($iProperty) as $link) {
			$iIndex++;
			$sId = self::formatAction($info->sName, $link->id);
			$this->pushElementLabel($sId, ucfirst($info->sAlone) . ' ' . $iIndex,
				self::getPropertyControls($sId, $aChoice, $link->idProperty),
				$page->linkEdit(new DenifLink(
					$page->getSelfTarget(targetParams, array('idDeleteLink' => $link->id)),
						'Enlever cette relation')));
		}
		return $iIndex;
	}
	
	/// @brief Add combo-boxes to the form to add links of the given type
	/// (\p $iProperty) to the current node.
	/// Each combo-box contains all properties of the given type.
	///
	/// Create <tt>aid..._add[]</tt> fields having strictly positive keys.
	/// All keys are created different.
	///
	/// @param $iProperty the type of the property.
	/// @param $nbRow the number of combo-boxes.
	public function pushAddLinks($iProperty, $nbRow)
	{
		if ($nbRow <= 0)
			return;
		
		$info = ObjectTypeInfo::get($iProperty);
		$aChoice = $this->getPropertyChoices(true, $iProperty);
		
		for ($iIndex = 1; $iIndex <= $nbRow; $iIndex++) {
			$sId = self::formatAction($info->sName, -$iIndex);
			$this->pushElementLabel($sId, 'Ajouter ' . $info->sUnUne,
				self::getPropertyControls($sId, $aChoice, 0),
				'Choisir ou laisser vide');
		}
	}
	
	/// @brief Add a text area containing all the links of the current node
	/// having a given type (\p $iProperty).
	/// Each combo-box contains all properties of the given type (\p $iProperty).
	///
	/// Create a <tt>s...[]</tt> field (the links list)
	/// and a <tt>i...</tt> combo-box that can add lines to the links list.
	/// The combo-box is only used locally, client-side, by the Javascript.
	/// The server does not use it at form submitting.
	///
	/// @param $iProperty the type of the property.
	public function pushSetLinks($iProperty)
	{
		$info = ObjectTypeInfo::get($iProperty);
		$aChoice = $this->getPropertyChoices(true, $iProperty);
		
		if ($this->isError())
			$sValues = $_REQUEST->getSafe('s' . $info->sName);
		else{
			$aValue = array();
			foreach ($this->obj->getPropertiesLink($iProperty) as $link)
				if ($link->idProperty > 0 && isset($aChoice[$link->idProperty]))
					$aValue[] = $aChoice[$link->idProperty];
			$sValues = concat($aValue, CRLF);
		}
		
		$page = $this->page;
		$sId = 's' . $info->sName;
		$this->pushElementLabel($sId, ucfirst($info->sPlural),
			form::textArea('ID' . $sId, -1, 4, html::text($sValues)),
			ucfirst($info->sPlural) . ' du nud,  raison dune ligne par ' . $info->sAlone);
		
		$sIdSelect = 'i' . $info->sName;
		$sScript = <<<EOF
var edt = this.form.elements["$sId"];
if (edt.value.length > 0 && edt.value.substr(edt.value.length - 1) != "\\n")
	edt.value += "\\n";
edt.value += this.options[this.selectedIndex].text;
EOF;
		$this->pushElementLabel($sIdSelect, 'Ajouter ' . $info->sUnUne,
			form::select('ID' . $sIdSelect, $aChoice, 0, array('onchange' => $sScript)));
	}
	
	/// @brief Add a line to the form to add a new property. This displays a new form.
	/// If it is validated, add a link to the new property to the current node.
	///
	/// @param $iProperty the type of the property that can be created.
	public function pushCreateProperty($iProperty)
	{
		$info = ObjectTypeInfo::get($iProperty);
		$this->pushElement('', form::submit(
			'bSubmitCreateProperty_' . $iProperty,
			'Crer ' . $info->sUnUne . ' puis lajouter'),
			'Ajouter ' . $info->sUnUne . ' qui nest pas dans la liste');
	}
	
	
	/// Return an array containing the main property types
	protected function getPropertyChoices($bAdd, $iProperty)
	{
		$aChoice = array();
		if ($bAdd)
			$aChoice[0] = '- Slectionnez ' . ObjectTypeInfo::get($iProperty)->sUnUne . ' -';
		foreach ($this->loadProperties($iProperty) as $property)
			$aChoice[$property->id] = $property->getLinkTitle();
		
		return html::textArray($aChoice);
	}
	
	protected static function getPropertyControls($sId, array $aChoice, $idProperty)
	{
		return form::select('ID' . $sId, $aChoice, $idProperty);
	}
	
	
	/// Return the form field ID of an action, by adding a prefix
	/// and eventually an bracketed array index to the field name.
	/// @param $sField the name of the action.
	/// @param $iIndex the index of the action, \c 0 to get the array name
	/// (no array brackets).
	public static function formatAction($sField, $iIndex = 0)
	{
		$s = 'aid' . $sField;
		if ($iIndex)
			$s .= '[' . $iIndex . ']';
		return $s;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Buttons
	/// @{
	//-----------------------------------------------------------------
	
	public function pushDelete()
	{
		$page = $this->page;
		if ($this->obj->canDelete())
			$this->pushElement('Supprimer', $page->linkEdit(
				getLinkDelete($page->obj, $page->targetDelete,
					$page->getSelfTarget(targetParams), 'Supprimer')));
	}
	
	public function pushSubmit(array $aButtonMore = array())
	{
		$aButton = array(
			form::submit('bSubmitReturn',   'Valider et retourner'),
			form::submit('bSubmitContinue', 'Valider et rester'),
		);
		append($aButton, $aButtonMore);
		$aButton[] = form::button('bCancel', 'Annuler',
			'location.href=' . html::jscriptString($this->page->targetExit->url()) . ';');
		
		$this->pushElement('', html::div(concat($aButton), attr::className('submits')));
	}
	
	/// @}
}


//-----------------------------------------------------------------
/// @name helper functions to build form-related links
/// @{
//-----------------------------------------------------------------

function getLinkDisabled($sAction, $sToolTip = '')
{
	return DenifLink::createAction('', $sAction . ' grayed', $sToolTip);
}

function getFormLink($sLabel, $sFormName, array $aParam, $sToolTip = '')
{
	static $aAction = array(
		'[edit]'      => 'edit',
		'[edit docs]' => 'edit-docs',
		'[delete]'    => 'delete',
		'[+]'         => 'add',
	);
	$sAction = defaultElement($aAction, $sLabel, '');
	
	$target = new target($sFormName . '.php', $aParam);
	return ($sAction)
		? DenifLink::createAction($target, $sAction, $sToolTip)
		: new DenifLink($target, $sLabel, '', $sToolTip);
}


function getLinkCreateBase($iObject, array $aParam, $sLabel)
{
	return getFormLink($sLabel,
		'edit_' . ((isDefineInRange($iObject, 'node')) ? 'node' : 'property'),
		$aParam, 'Ajouter ' . ObjectTypeInfo::get($iObject)->sUnUne);
}


function getLinkCreate($iObject, target $targetExit, $idParent = 0)
{
	$aParam = array(
		'iObject'  => $iObject,
		'sExitURL' => PageDenif::targetGetCleanURL($targetExit),
	);
	if ($idParent)
		$aParam['idParent'] = $idParent;
	
	return getLinkCreateBase($iObject, $aParam,
		'Ajouter ' . ObjectTypeInfo::get($iObject)->sUnUne);
}


/// <tt>$iOrder \< 0</tt> to insert the node at the end.
function getLinkCreateChild($iNodeChild, $idParent, $iOrder, target $targetExit, $sLabel = '[+]')
{
	return getLinkCreateBase($iNodeChild,
		array(
			'iObject'  => $iNodeChild,
			'sExitURL' => PageDenif::targetGetCleanURL($targetExit),
			'idParent' => $idParent,
			'iOrder'   => $iOrder,
		), $sLabel);
}


append(DenifLink::$aActionToolTip, array(
	'edit'      => 'diter',
	'edit-docs' => 'diter les documents',
	'delete'    => 'Supprimer',
	'add'       => 'Ajouter',
	'up'        => 'Monter',
	'down'      => 'Descendre',
));

function getLinkEdit(ObjectBase $obj, target $targetExit, $sLabel = '[edit]')
{
	return getFormLink($sLabel, 'edit_' . $obj->getFormName(),
		array(
			'iObject'  => $obj->getType(),
			'id'       => $obj->id,
			'sExitURL' => PageDenif::targetGetCleanURL($targetExit),
		), 'diter ' . $obj->getInfo()->sLeLa);
}

function getLinkEditDocs(node $node, target $targetExit, $sLabel = '[edit docs]')
{
	return getFormLink($sLabel, 'edit_docs',
		array(
			'id'       => $node->id,
			'sExitURL' => PageDenif::targetGetCleanURL($targetExit),
		), 'diter les documents');
}


function getLinkMove(node $node, $iOffset, $iOrderLast, $target)
{
	if ($iOffset < 0) {
		$sAction  = 'up';
		$bEnabled = ($node->iOrder > 0);
	}else{
		$sAction  = 'down';
		$bEnabled = ($node->iOrder < $iOrderLast);
	}
	if (!$bEnabled)
		return getLinkDisabled($sAction);
	
	$target->addParams(array(
		'iMove'     => $iOffset,
		'idMove'    => $node->id,
		'iMoveFrom' => $node->iOrder,
	));
	return DenifLink::createAction($target, $sAction);
}


function getLinkDelete(ObjectBase $obj, target $targetConfirm, target $targetCancel, $sLabel = '[delete]')
{
	return getFormLink($sLabel, 'delete',
		array(
			'iObject'     => $obj->getType(),
			'id'          => $obj->id,
			'sConfirmURL' => PageDenif::targetGetCleanURL($targetConfirm),
			'sCancelURL'  => PageDenif::targetGetCleanURL($targetCancel),
		));
}

///@}

?>