<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder

//-----------------------------------------------------------------
/// @file
/// @brief #JSON: formatter to exchange data is a JScript-like format,
/// see http://www.json.org for details
//-----------------------------------------------------------------


// Copyright (c) 2006 - 2007 Andrea Giammarchi
//
// Permission is hereby granted, free of charge,
// to any person obtaining a copy of this software and associated
// documentation files (the "Software"),
// to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


/// JSON parser and writer.
class JSON
{
	/// High-level function: prints a value in JSON format.
	public static function output($value)
	{
		useMimeText();
		echo utf8::encode(self::encode($value));
		exit;
	}
	
	/// High-level function: decode a JSON string.
	public static function input($s)
	{
		return self::decode(utf8::decode($s));
	}
	
	/// High-level function: decode a JSON string to a safe object (ObjectSafe).
	public static function inputSafe($s)
	{
		return new ObjectSafe(self::input($s));
	}
	
	/// Low-level function: decode a JSON string.
	public static function decode($s)
	{
		$pos = 0;
		return (is_string($s))
			? self::decodeRecurse($s, $pos)
			: null;
	}
	
	/// Low-level function: encode a value to JSON format.
	public static function encode($value)
	{
		switch (gettype($value)) {
			case 'boolean':
				return ($value) ? 'true' : 'false';
			
			case 'integer':
			case 'double':
			case 'float':
				return strval($value);
			
			case 'string':
				return self::encodeString($value);
			
			case 'array':
				if (!count($value) || array_keys($value) === range(0, count($value) - 1)) {
					$a = map(array(__CLASS__, 'encode'), $value);
					return '[' . concat($a, ',') . ']';
				}
				$a = array();
				foreach ($value as $key => $element)
					if (($element = self::encode($element)) !== '') {
						if (!is_integer($key) || $key < 0)
							$key = self::encode(strval($key));
						$a[] = $key . ':' . $element;
					}
				return '{' . concat($a, ',') . '}';
			
			case 'object':
				$a = array();
				foreach (get_object_vars($value) as $key => $element)
					if (($element = self::encode($element)) !== '')
						$a[] = self::encode($key) . ':' . $element;
				return '{' . concat($a, ',') . '}';
			
			default:
				return 'null';
		}
	}
	
	
	private static function convertString($s, $iFrom, $iTo)
	{
		static $aReplace = null;
		if (!$aReplace) {
			foreach (array_merge(range(0, 7), array(11), range(14, 31)) as $i) {
				$aReplace[0][] = chr($i);
				$aReplace[1][] = '\u00' . sprintf('%02x', $i);
			}
			append($aReplace[0], map('chr', array(0x5C, 0xFF, 0x22, 0x0D, 0x0C, 0x0A, 0x09, 0x08)));
			append($aReplace[1], getList('\\\\;\\/;\\";\r;\f;\n;\t;\b'));
		}
		
		return str_replace($aReplace[$iFrom], $aReplace[$iTo], $s);
	}
	
	
	public static function encodeString($s)
	{
		return '"' . self::convertString($s, 0, 1) . '"';
	}
	
	
	private static function decodeRecurse($s, &$pos)
	{
		switch ($s[$pos]) {
			case 't':
				$pos += 4;
				return true;
			
			case 'n':
				$pos += 4;
				return null;
			
			case 'f':
				$pos += 5;
				return false;
			
			case '[':
				$pos++;
				$a = array();
				while ($s[$pos] !== ']') {
					$a[] = self::decodeRecurse($s, $pos);
					if ($s[$pos] === ',')
						$pos++;
				}
				$pos++;
				return $a;
			
			case '{':
				$a = array();
				$pos++;
				while ($s[$pos] !== '}') {
					$index = self::decodeString($s, $pos);
					$pos++;
					$a[$index] = self::decodeRecurse($s, $pos);
					if ($s[$pos] === ',')
						$pos++;
				}
				$pos++;
				return $a;
			
			case '"':
				return self::decodeString($s, $pos);
			
			default:
				$sResult = '';
				preg_replace('/^(\-)?([0-9]+)(\.[0-9]+)?([eE][+-]?[0-9]+)?/e', '$sResult = "\\1\\2\\3\\4"', substr($s, $pos));
				if ($sResult === '')
					return '';
				$pos += strlen($sResult);
				$valueInt   = intval($sResult);
				$valueFloat = floatval($sResult);
				return ($valueInt == $valueFloat) ? $valueInt : $valueFloat;
		}
	}
	
	private static function decodeString($s, &$pos)
	{
		$pos2 = $pos++;
		do{
			$pos2 = strpos($s, '"', $pos2 + 1);
			if ($pos2 === false)
				die;
		}while (self::isEscaped($s, $pos2));
		$lenString = $pos2 - $pos;
		
		$s = self::convertString(substr($s, $pos, $lenString), 1, 0);
		$pos += $lenString + 1;
		return $s;
	}
	
	private static function isEscaped($s, $pos)
	{
		$nbBackslash = 0;
		while ($s[--$pos] === '\\')
			$nbBackslash++;
		return $nbBackslash % 2;
	}
}

?>