<?php
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder

//-----------------------------------------------------------------
/// @file
/// @brief Global definitions used by all other files
//-----------------------------------------------------------------


/// @brief Defined when <tt>kernel.inc.php</tt> is included.
/// Used by library files to display fatal error
/// when the kernel is not available.
define('KERNEL_INCLUDED', true);

/// Timestamp of the beginning of PHP execution.
$g_dtBegin = microtime(true);


//-----------------------------------------------------------------
/// @name Default settings
/// @{
//-----------------------------------------------------------------

/// @brief Position of this file (<tt>kernel.inc.php</tt>)
/// in the directories tree, relative to the website root.
///
/// If <tt>kernel.inc.php</tt> is at the root (default assumption), give 0.
defineDefault('iKernelDirLevel', 0);

/// @brief Print debug messages: errors, tracing information.
defineDefault('bDebug', false);

/// @brief Absolute filename of the PHP error log file.
/// \c null to disable error logging.
defineDefault('sLogFile', null);

/// @brief Maximum size of the errors log file. Used only if #sLogFile is defined.
defineDefault('maxSizeLogFile', 100 << 10);

/// @brief Print debugging information about the cache system.
/// Useful if you notice the cached files are not updated when necessary.
defineDefault('bDebugCache', bDebug && false);

/// @brief Directory of the website root relative to the currently displayed file.
defineDefault('sRootRel', str_repeat('../',
	countDirLevels(realpath($_SERVER['SCRIPT_FILENAME'])) - (countDirLevels(__FILE__) - iKernelDirLevel)));

/// @brief Encoding of the generated XHTML file. Used by the default DOCTYPE,
/// by Page::head() to generate the Content-Type <tt>\<meta\></tt> tag,
/// and by XML writing functions to escape text.
defineDefault('sEncoding', 'windows-1252');

/// @brief Header of the generated XHTML file. You may want to use another encoding.
defineDefault('sDocType',
	'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">');

/// @brief MIME type used for unknown file extensions.
/// Give \c null to use no default MIME type.
defineDefault('sDefaultMimeType', 'text/plain');

/// @brief String used as newline and inserted before and/or after some XML tags,
/// for readability purpose. Empty by default (no newline is inserted).
/// Set it to <tt>"\n"</tt> or <tt>"\r\n"</tt> to enable newlines writing.
defineDefault('sXMLNewline', '');

/// @brief Change directory to the executed script directory. Enabled by default.
defineDefault('bAutoChdir', true);


/// Used to translate the result of some display functions.
$g_bFrench = false;

iconv_set_encoding('input_encoding',    sEncoding);
iconv_set_encoding('internal_encoding', sEncoding);
iconv_set_encoding('output_encoding',   sEncoding);

if (bAutoChdir)
	chdir(dirname($_SERVER['SCRIPT_FILENAME']));

/// @}


//-----------------------------------------------------------------
/// @name Constants
/// @{
//-----------------------------------------------------------------

/// \c true if PHP runs under Windows, \c false otherwise.
define('bWindows',   strtoupper(substr(PHP_OS, 0, 3)) == 'WIN');

/// @brief \c true if the script is launched from command line,
/// \c false if it is launched from a Web server.
define('bCmdLine',   defined('STDIN'));

/// Mac end-of-line string: <tt>"\r"</tt>
define('CR',         "\r");

/// Unix end-of-line string: <tt>"\n"</tt>
define('LF',         "\n");

/// Windows end-of-line string: <tt>"\r\n"</tt>
define('CRLF',       "\r\n");

/// Two Unix end-of-line characters: <tt>"\n\n"</tt>
define('LFLF',       "\n\n");

/// Tabulation character: <tt>"\t"</tt>
define('TAB',        "\t");

/// Basename of the running script.
define('sSelfName',  basename($_SERVER['PHP_SELF']));

/// Reasonably good regexp for email addresses, to be used by <tt>eregi()</tt>.
define('sEmailRegexp', "^([-!#\$%&'*+./0-9=?A-Z^_`a-z{|}~])+@([-!#\$%&'*+/0-9=?A-Z^_`a-z{|}~]+\\.)+[a-zA-Z]{2,6}\$");

/// @}


//-----------------------------------------------------------------
/// @name Constant definition helpers
/// @{
//-----------------------------------------------------------------

/// Define a constant, only if it not defined yet.
function defineDefault($sConstant, $defaultValue)
{
	if (!defined($sConstant))
		define($sConstant, $defaultValue);
}

/// Define a list of constants from a <tt>'name' => 'value'</tt> array.
function defineArray(array $a)
{
	iterKeyValue('define', $a);
}

/// @brief Define an enumeration of constants,
/// whose values are from 0 to N-1.
///
/// @param $sPrefix common prefix to all constants
/// @param $sList \c ';' separated list of the constants names
/// @param $bCount \c true to define <tt>$sPrefix.'First'</tt>,
/// <tt>$sPrefix.'Last'</tt> and <tt>$sPrefix.'Count'</tt> (whose value is N).
/// @param $iFirst value of the first element.
function defineEnum($sPrefix, $sList, $bCount = false, $iFirst = 0)
{
	$iCount = 0;
	foreach (getList($sList) as $sName)
		define($sPrefix . $sName, $iFirst + $iCount++);
	if ($bCount) {
		define($sPrefix . 'First', $iFirst);
		define($sPrefix . 'Last',  $iFirst + $iCount - 1);
		define($sPrefix . 'Count', $iCount);
	}
}

/// Indicate whether an enum value is in the range defined by <tt>defineEnum(bCount=true)</tt>.
function isDefineInRange($iValue, $sPrefix)
{
	return constant($sPrefix . 'First') <= $iValue && $iValue <= constant($sPrefix . 'Last');
}

function getDefineRange($sPrefix)
{
	return range_incr(constant($sPrefix . 'First'), constant($sPrefix . 'Last'));
}

/// @brief Define a list of flags constants,
/// whose values are from <tt>1 \<\< 0</tt> to <tt>1 \<\< (N-1)</tt>.
///
/// @param $sPrefix common prefix to all constants
/// @param $sList ';' separated list of the constants names
function defineFlags($sPrefix, $sList)
{
	$index = 0;
	foreach (getList($sList) as $sName)
		define($sPrefix . $sName, 1 << ($index++));
}

/// @}


//-----------------------------------------------------------------
/// @name Functionnal API
/// @{
//-----------------------------------------------------------------

/// @brief Apply a function to each element of an array.
/// The callback prototype is: <tt>($value) -\> $result</tt>.
function map($callback, array $a)
{
	return array_map($callback, $a);
}

/// @brief Map a function to all keys and values of an array.
/// The callback prototype is: <tt>($key, $value) -> $result</tt>
function mapKeyValue($callback, array $a)
{
	return array_map($callback, array_keys($a), array_values($a));
}

/// @brief Map a function to all keys and values of an array.
/// The callback prototype is: <tt>($value, $key) -> $result</tt>
function mapValueKey($callback, array $a)
{
	return array_map($callback, array_values($a), array_keys($a));
}

/// @brief Map a function to all keys of an array.
/// The callback prototype is: <tt>($key) -> $result</tt>.
function mapKey($callback, array $a)
{
	return array_map($callback, array_keys($a));
}


/// @brief Map a function to all elements of an array, then concat their string result.
/// The callback prototype is: <tt>($value) -> $result</tt>
function mapConcat($callback, array $a, $sSeparator = '')
{
	return concat(array_map($callback, $a), $sSeparator);
}

/// @brief Map a function to all elements of an array, then concat their string result.
/// The callback prototype is: <tt>($key, $value) -> $string</tt>
function mapConcatKeyValue($function, array $a, $sSeparator = '')
{
	return concat(mapKeyValue($function, $a), $sSeparator);
}

/// @brief Map a function to all keys of an array, then concat their string result.
/// The callback prototype is: <tt>($key) -> $string</tt>
function mapConcatKey($function, array $a, $sSeparator = '')
{
	return concat(mapKey($function, $a), $sSeparator);
}

/// @brief Execute a function for each element of an array.
/// The callback prototype is: <tt>($value) -> void</tt>
function iter($callback, array &$a)
{
	array_walk($a, $callback);
}

/// @brief Execute a function for each element of an array.
/// The callback prototype is: <tt>($key, $value) -> void</tt>
function iterKeyValue($callback, array &$a)
{
	foreach ($a as $key => &$value)
		call_user_func($callback, $key, $value);
}

/// @brief Execute a function for each element of an array.
/// The callback prototype is: <tt>($value, $key) -> void</tt>
function iterValueKey($callback, array &$a)
{
	array_walk($a, $callback);
}


/// @brief Create a temporary function and return its name.

/// @param $sArgs ';' separated list of the arguments names
/// @param $sCode PHP code of the function
function fun($sArgs, $sCode)
{
	$aArg = array();
	foreach (getList($sArgs) as $sName)
		$aArg[] = '$' . $sName;
	return create_function(implode(',', $aArg), $sCode);
}


/// @brief Wrapper for str_replace,
/// taking one associative array instead of two arrays
function replace($sText, array $aReplace)
{
	return str_replace(array_keys($aReplace), array_values($aReplace), $sText);
}


/// @brief Projection: returns an array of objects
/// having as only member the one named \c $sMember.
function projectTo(array $aObject, $sMember)
{
	$aMember = array();
	foreach ($aObject as $obj)
		$aMember[] = $obj->$sMember;
	return $aMember;
}

/// @brief projectTo + array_unique: all elements of the returned array are unique.
/// @sa projectTo
function projectToUnique(array $aObject, $sMember)
{
	return array_unique(projectTo($aObject, $sMember));
}

function projectToId(array $aObject)
{
	return projectTo($aObject, 'id');
}


/// @brief Projection of an array of objects to one of the objects member,
/// preserving the key.
function projectToKey(array $aObject, $sMember)
{
	$aMember = array();
	foreach ($aObject as $key => $obj)
		$aMember[$key] = $obj->$sMember;
	return $aMember;
}


/// Get all objects of an array having the given value in a member.
function filterBy(array $aObject, $sMember, $value)
{
	$aResult = array();
	foreach ($aObject as $obj)
		if ($obj->$sMember === $value)
			$aResult[] = $obj;
	return $aResult;
}


/// @brief Return an array with each \c $obj stored under key <tt>$obj->$sMemberKey</tt>,
/// i.e. an array associating to each key the only object matching this key.
function bijectiveBy(array $aObject, $sMemberKey)
{
	$aResult = array();
	foreach ($aObject as $obj)
		$aResult[$obj->$sMemberKey] = $obj;
	return $aResult;
}


/// Return an array associating to each key the list of objects matching this key.
function assocBy(array $aObject, $sMemberKey)
{
	$aResult = array();
	foreach ($aObject as $obj)
		$aResult[$obj->$sMemberKey][] = $obj;
	return $aResult;
}


/// @brief Return an array associating to each <tt>$obj->$sMemberKey</tt>
/// the list of <tt>$obj->$sMemberValue</tt> attached to the \c $obj
/// matching this <tt>$obj->$sMemberKey</tt>.
function assocFieldBy(array $aObject, $sMemberKey, $sMemberValue)
{
	$aResult = array();
	foreach ($aObject as $obj)
		$aResult[$obj->$sMemberKey][] = $obj->$sMemberValue;
	return $aResult;
}

/// @}


//-----------------------------------------------------------------
/// @name String & array helpers
/// @{
//-----------------------------------------------------------------

/// @brief Indicate if a string is \c null or empty.
function isEmpty($s)
{
	return is_null($s) || $s === '';
}

/// @brief Indicate if a string not \c null nor empty.
function isSome($s)
{
	return !is_null($s) && $s !== '';
}

/// @brief Indicate if a string has the given prefix.
function hasPrefix($s, $sPrefix)
{
	return @substr_compare($s, $sPrefix, 0, strlen($sPrefix)) === 0;
}


/// @brief Restrict a value to a \p $min ... \p $max interval.
/// Undefined behavior if <tt>\p $min \> \p $max</tt>.
function minmax($value, $min, $max)
{
	return max($min, min($value, $max));
}


/// @brief <tt>($value) ? $value : $defaultValue</tt>
function defaultValue($value, $defaultValue)
{
	return ($value) ? $value : $defaultValue;
}

/// @brief Return the value associated to a key in an array,
/// or the given value if the key is not defined in the array.
function defaultElement(array $a, $key, $defaultValue)
{
	return (array_key_exists($key, $a)) ? $a[$key] : $defaultValue;
}


/// Concatenate all strings of an array.
function concat($a, $sSeparator = '')
{
	return ($a) ? implode($sSeparator, $a) : '';
}

/// Create an array associating keys to values.
function combine($aKey, $aValue)
{
	return ($aKey) ? array_combine($aKey, $aValue) : array();
}

/// Concatenate all non-empty strings of an array.
function concatFilter($a, $sSeparator = '')
{
	return ($a) ? implode($sSeparator, array_filter($a)) : '';
}

/// Convert a \c ';' separated list to an array.
function getList($s)
{
	return ($s) ? explode(';', $s) : array();
}

/// Convert a \c ';' separated list to an array, if necessary.
function getListSafe($list)
{
	return (is_array($list))
		? $list
		: getList($list);
}

/// Convert an array to a \c ';' separated list.
function toList(array $a)
{
	return implode(';', $a);
}

/// Concat any number of \c ';' separated lists.
function concatLists()
{
	$aAll = func_get_args();
	return concat(';', array_filter($aAll));
}


/// @brief Put some arrays of equal length and keys side-by-side.
/// <tt>array_concatcols(array(1, 2, 3), array(4, 5, 6)) = array(array(1, 4), array(2, 5), array(3, 6))</tt>
function concatVectors()
{
	$aAll = func_get_args();
	return array_map(null, $aAll);
}


/// @brief Transform an <tt>array(key1 => value1, ..., keyN => valueN)</tt>
/// into an <tt>array(array(key1, value1), ..., array(keyN, valueN))</tt>
function concatKeyVal(array $a)
{
	return array_map(null, array_keys($a), array_values($a));
}


/// @brief Return the first element of an array, or \c false if the array is empty.
function firstElement(array $a)
{
	return reset($a);
}

/// @brief Remove duplicates in an array, by side effect.
function makeUnique(array &$a)
{
	$a = array_unique($a);
}

/// @brief Append some values to an array, by side effect.
/// The values of the appended array will overwrite the values of the modified array.
function append(array &$a, array $aMore)
{
	$a = array_merge($a, $aMore);
}

/// @brief Add values to an array, by side effect.
/// The values of the appended array will never overwrite the values of the modified array.
function appendDefault(array &$a, array $aMore)
{
	$a = array_merge($aMore, $a);
}

/// @brief Concatenate all arrays and single values contained in an array to a single array.
function flatten(array $a)
{
	$aResult = array();
	flattenInternal($aResult, $a);
	return $aResult;
}

/// @cond Internal function
function flattenInternal(array &$aResult, array $a)
{
	foreach ($a as $value)
		if (is_array($value))
			flattenInternal($aResult, $value);
		else
			$aResult[] = $value;
}
/// @endcond

/// @brief Convert object \p $obj to an array, including only fields \p $aField.
/// Set <tt>$aField = array('*')</tt> to get all fields.
function getObjectFields($obj, array $aField = array('*'))
{
	if ($aField == array('*'))
		return (array)$obj;
	
	$a = array();
	foreach ($aField as $sField)
		$a[$sField] = $obj->$sField;
	return $a;
}

/// Return <tt>'s'</tt> or <tt>''</tt>, depending of \p $value.
function plural($value)
{
	$value = (is_array($value)) ? count($value) : abs($value);
	return ($value >= 2) ? 's' : '';
}

/// @brief Append a plural suffix to a string, depending of the number \p $value.
/// <tt>pluralString(42, 'foo')</tt> returns <tt>"42 foos"</tt>.
function pluralString($value, $s)
{
	return $value . ' ' . $s . plural($value);
}


/// @brief Create an array indexed by an integer range.
/// If <tt>$iFirst \> $iLast</tt>, return an empty array.
function array_fill_range($iFirst, $iLast, $value)
{
	return ($iFirst <= $iLast)
		? array_fill($iFirst, $iLast - $iFirst + 1, $value)
		: array();
}

/// @brief Same as <tt>range()</tt>,
/// but return an empty array if <tt>$iFirst \> $iLast</tt>.
function range_incr($iFirst, $iLast)
{
	return ($iFirst <= $iLast) ? range($iFirst, $iLast) : array();
}

/// @}


//-----------------------------------------------------------------
/// @name Type deduction from name prefix
/// @{
//-----------------------------------------------------------------

/// Deduce the type of a variable from its name prefix.
function deduceType($sName)
{
	if (!preg_match('/^(i|nb|dt|s|b|a)/', $sName, $aMatch))
		return null;
	
	static $s_aAssoc = array(
		'i'  => 'integer',
		'nb' => 'integer',
		'dt' => 'integer',
		's'  => 'string',
		'b'  => 'boolean',
		'a'  => 'array',
	);
	
	return $s_aAssoc[$aMatch[1]];
}


/// Force the type of a variable contents by deducing it from its name prefix.
function forceType(&$var, $sName)
{
	$sType = deduceType($sName);
	
	if ($sType == 'array') {
		if (empty($var) || !is_array($var))
			$var = array();
		else{
			$sType = deduceType(substr($sName, 1));
			if ($sType)
				foreach ($var as &$value)
					settype($value, $sType);
		}
		
	}else if ($sType) {
		if (is_array($var))
			$var = null;
		@settype($var, $sType);
	}
}

/// Force the type of the given members of an object.
function forceTypeMembers(&$obj, $sMembers)
{
	foreach (getList($sMembers) as $sMember) {
		if (!isset($obj->$sMember))
			$obj->$sMember = null;
		forceType($obj->$sMember, $sMember);
	}
}


/// @brief Create a string-indexed array from an integer-indexed array.
/// The result array takes its keys from \p $sKeys, and its values from \p $a.
/// The values are converted according to the prefix of their key,
/// with the forceType() function.
///
/// The keys of \p $a are ignored.
/// If \p $a has more elements than $sKeys, the extra elements are lost.
/// If it has less, the result array will use empty values to the missing values.
/// \code
/// giveArrayKeys(array(1, 2), 'iA;sB;sC') = array('iA' => 1, 'sB' => '2', 'sC' => '')
/// giveArrayKeys(array(1, 2, 3), 'iA;iB') = array('iA' => 1, 'iB' => 2)
/// \endcode
///
/// @param $a the array to convert. It remains unchanged.
/// @param $sKeys the list of the keys name.
/// @return an array having exactly the keys given by \p $sKeys,
/// and values taken from \p $a (or empty values) converted by forceType().
function giveArrayKeys(array $a, $sKeys)
{
	$aResult = array();
	reset($a);
	foreach (getList($sKeys) as $sKey) {
		list($key, $value) = each($a);
		forceType($sKey, $value);
		$aResult[$sKey] = $value;
	}
	return $aResult;
}


/// @}


//-----------------------------------------------------------------
/// @brief Static class taking care of automatic classes inclusion.
/// Included files should be in the same directory than this file or in #sRootRel,
/// and should have a <tt>'.inc.php'</tt> extension.
//-----------------------------------------------------------------

class ClassLoader
{
	/// @brief Class-to-file associations, for automatic class inclusion
	/// in case where the file has not the same name than the file.
	/// Keys are classes name, in lowercase; values are files basename.
	///
	/// If class \c MyClass needs to be autoloaded and is defined in file \c MyFile
	/// where <tt>MyClass != MyFile</tt>, add <tt>'MyClass' => 'MyFile'</tt> to this array.
	private static $aAlias = array();
	
	
	/// @brief Register, in the alias table, the file that defines a class.
	public static function registerAlias($sClass, $sFileBasename)
	{
		self::$aAlias[strtolower($sClass)] = $sFileBasename;
	}
	
	/// @brief Register several aliases.
	public static function registerAliases(array $aClassToFileBasename)
	{
		iterKeyValue(array(__CLASS__, 'registerAlias'), $aClassToFileBasename);
	}
	
	
	/// @brief Return the full path of a file belonging
	/// to the directory that contains <tt>kernel.inc.php</tt>
	/// or to the directory that contains the currently executed PHP file.
	/// The returned path may not point to an existing file.
	public static function getFullPath($sBasename)
	{
		$s = dirname(__FILE__) . '/' . $sBasename;
		if (!is_file($s)) {
			$s = (bAutoChdir)
				? ''
				: dirname($_SERVER->SCRIPT_FILENAME) . '/';
			$s .= sRootRel . $sBasename;
		}
		
		return $s;
	}
	
	
	/// @brief Load a PHP include file contained in the directory that contains
	/// <tt>kernel.inc.php</tt>.
	///
	/// @param $sBasename the name of the file to load, without <tt>.inc.php</tt> extension.
	/// @param $bRequireSuccess if \c true and the file to load is not found,
	/// exit with an error message. If \c false, return \c true on success, \c false on error.
	public static function loadFile($sBasename, $bRequireSuccess = true)
	{
		$sPath = self::getFullPath($sBasename . '.inc.php');
		
		if (!is_file($sPath)) {
			if ($bRequireSuccess)
				die('ClassLoader::loadFile: file not found: ' . $sPath);
			return false;
		}
		
		require_once $sPath;
		return true;
	}
	
	
	/// @brief Load the PHP file of a class. Use #$aAlias to revolve aliases,
	/// then load the PHP file having the same name than the class,
	/// but in lowercase and with <tt>.inc.php</tt> extension.
	public static function loadClass($sClass)
	{
		$sBasename = strtolower($sClass);
		$sBasename = defaultElement(self::$aAlias, $sBasename, $sBasename);
		self::loadFile($sBasename, false);
	}
}

/// @brief Function called when PHP needs to load a class:
/// delegate this task to ClassLoader::loadClass().
function __autoload($sClass)
{
	ClassLoader::loadClass($sClass);
}


ClassLoader::registerAliases(array(
	'BrowseDir' => 'path',
	'target'    => 'links',
	'link'      => 'links',
));

ClassLoader::loadFile('html');

debug::setup();

/// @}


//-----------------------------------------------------------------
/// @name Various helpers
/// @{
//-----------------------------------------------------------------

/// Return the depth level of a path, 0 if the file is at root.
function countDirLevels($sPath)
{
	return substr_count($sPath, '/') + substr_count($sPath, '\\');
}


/// Convert a boolean to a 0/1 integer.
function toInt($b)
{
	return ($b) ? 1 : 0;
}


/// Formatted time elapsed since PHP started.
function elapsedTime()
{
	global $g_bFrench, $g_dtBegin;
	
	$s = sprintf('%.3f s', microtime(true) - $g_dtBegin);
	return ($g_bFrench)
		? str_replace('.', ',', $s)
		: $s;
}


/// @brief Detect the MIME type of a file from its file extension.
/// Use the 'mime.txt' library file, not the builtin function mime_content_type
function guessMimeType($sFile)
{
	$sExt = strtolower(path::getExtension($sFile));
	if ($sExt == '')
		return sDefaultMimeType;
	
	static $aMime = null;
	if (!$aMime) {
		$aMime = array();
		foreach (file(getFullPath('mime.txt')) as $sLine) {
			sscanf($sLine, '%s %s', $sMimeExt, $sMimeType);
			$aMime[$sMimeExt] = $sMimeType;
		}
	}
	
	return defaultElement($aMime, $sExt, sDefaultMimeType);
}


/// @brief Obfuscate an email address.
/// Convert dots and <tt>\@</tt> symbols to text, obfuscate text using ROT13.
/// The returned text is intended to be unobfuscated by unobfuscateMail() Javascript function.
function obfuscateMail($sEmail)
{
	return str_rot13(replace($sEmail, array(
		'@' => '-at-',
		'.' => '-dot-',
	)));
}


/// @brief Run a performance test by executing a PHP code fragment many times.
/// Print results as HTML code.
function benchmark($sPHPCode, $nbIteration = 1000)
{
	$timeBegin = microtime(true);
	eval('for ($i = 0; $i < ' . $nbIteration . '; $i++) { ' . $sPHPCode . ' }');
	$timeEnd = microtime(true);
	$timeDiff = $timeEnd - $timeBegin;
	echo html::par(
		'Benchmark: ' . html::pre(html::text($sPHPCode)) .
		sprintf('Total time: %.3f', $timeDiff) . html::br() .
		sprintf('Average iteration time: %.3f', $timeDiff / $nbIteration));
}

/// @}


//-----------------------------------------------------------------
/// @brief Static class for end-of-line conversion
//-----------------------------------------------------------------

class endofline
{
	const sRegexp = '(\\r\\n|\\r|\\n)';
	const sWin  = "\r\n";
	const sUnix = "\n";
	const sMac  = "\r";
	
	
	/// @brief Convert the end-of-lines of a string to Windows format: CRLF.
	public static function toWin($s)
	{
		return self::toCustom($s, self::sWin);
	}
	
	/// @brief Convert the end-of-lines of a string to Unix format: LF.
	public static function toUnix($s)
	{
		return self::toCustom($s, self::sUnix);
	}
	
	/// @brief Convert the end-of-lines of a string to Mac format: LF.
	public static function toMac($s)
	{
		return self::toCustom($s, self::sMac);
	}
	
	/// @brief Replace the end-of-lines of a string by <tt>\<br /\></tt> tags.
	public static function toBR($s)
	{
		return regexp::replace(self::sRegexp, html::br(), $s);
	}
	
	/// @brief Replace the end-of-lines of a string by another text.
	public static function toCustom($s, $sReplaceBy)
	{
		return regexp::replace(self::sRegexp, $sReplaceBy, $s);
	}
}



/// @cond Function definitions within if/then/else confuse Doxygen

//-----------------------------------------------------------------
/// @name Magic quotes removing
/// @{
//-----------------------------------------------------------------

if (get_magic_quotes_gpc()) {
	
	function stripMagicQuotesValue(&$value)
	{
		$value = stripslashes($value);
	}
	
	function stripMagicQuotesArray(array &$a)
	{
		array_walk_recursive($a, 'stripMagicQuotesValue');
	}
	
	function stripMagicQuotes(&$value)
	{
		if (!empty($value)) {
			if (is_array($value))
				stripMagicQuotesArray($value);
			else
				stripMagicQuotesValue($value);
		}
	}
	
}else{
	function stripMagicQuotesValue(&$value) {}
	function stripMagicQuotesArray(array &$a) {}
	function stripMagicQuotes(&$value) {}
}

/// @}
/// @endcond



//-----------------------------------------------------------------
/// Static class for converting between UTF-8 and #sEncoding
//-----------------------------------------------------------------

class utf8
{
	public static function encode($s)
	{
		return iconv(sEncoding, 'UTF-8//TRANSLIT', $s);
	}
	
	public static function decode($s)
	{
		return iconv('UTF-8', sEncoding . '//TRANSLIT', $s);
	}
	
	public static function encodeRecurse($value)
	{
		return (is_array($value))
			? map(array(__CLASS__, __FUNCTION__), $value)
			: self::encode($value);
	}
	
	public static function decodeRecurse($value)
	{
		return (is_array($value))
			? map(array(__CLASS__, __FUNCTION__), $value)
			: self::decode($value);
	}
}



//-----------------------------------------------------------------
/// @brief Class wrapper around auto-global arrays.
///
/// Allows to consider \c $_REQUEST, \c $_POST, etc. as objects
/// having properties.
///
/// The 'prepare' and 'safe' methods convert the variables
/// to a safe format according the following rules:
/// - type: it is set by the forceType() function,
///   according to the variable prefix
/// - default value: if the variable is null,
///   it is given the default value of its type
/// - magic quotes: if the array is concerned by magic quotes,
///   quotes are stripped
///
/// Typical usage:
/// \code
/// $_REQUEST->prepare('sString;bBoolean;iInteger');
/// echo "A string: $_REQUEST->sString<br />";
/// echo "An integer: $_REQUEST->iInteger<br />";
/// extract($_REQUEST->getSafeList('sString;bBoolean;iInteger');
/// echo "A string: $sString<br />";
/// echo "An integer: $iInteger<br />";
/// \endcode
//-----------------------------------------------------------------

class ArrayWrapper implements Iterator
{
	private $array;
	
	
	public function __construct(array &$array, $removeUseMagicQuotes = false)
	{
		$this->array = &$array;
		if ($removeUseMagicQuotes)
			stripMagicQuotesArray($this->array);
	}
	
	/// @brief Return the value of a variable in the array.
	/// Generate a warning if the variable is not set.
	/// The ArrayWrapper::prepare() method should be called before
	/// to force the type of the variable.
	public function __get($sKey) { return $this->array[$sKey]; }

	/// Modify a variable of the array.
	public function __set($sKey, $value) { $this->array[$sKey] = $value; }
	
	/// Indicate whether a variable is defined in the array.
	public function __isset($sKey) { return isset($this->array[$sKey]); }
	
	/// Remove a value.
	public function __unset($sKey) { unset($this->array[$sKey]); }
	
	
	/// @brief Return the value of a variable in the array.
	/// If the variable is not set, return the given default value.
	public function get($sKey, $defaultValue = null)
	{
		return (isset($this->array[$sKey]))
			? $this->array[$sKey]
			: $defaultValue;
	}
	
	/// Give a default value for a variable in the array.
	public function setDefault($sKey, $defaultValue)
	{
		if (!isset($this->array[$sKey]))
			$this->array[$sKey] = $defaultValue;
	}
	
	/// @brief Return the value of a variable in the array converted to the given type.
	/// The type name follows the \c settype() function specification.
	public function& getAs($sKey, $sType)
	{
		settype($this->array[$sKey], $sType);
		return $this->array[$sKey];
	}
	
	
	/// Convert a variable of the array to safe format then return it.
	/// @see ArrayWrapper
	public function& getSafe($sKey)
	{
		$this->prepareOne($sKey);
		return $this->array[$sKey];
	}
	
	/// Same as getSafe, but convert from UTF8.
	public function getSafeUTF8($sKey)
	{
		return utf8::decodeRecurse($this->getSafe($sKey));
	}
	
	
	/// @brief Return an array containing a list of variables
	/// converted to safe format and indexed by their name.
	/// @see ArrayWrapper
	public function getSafeList($sKeys)
	{
		$a = array();
		foreach (getList($sKeys) as $sKey)
			$a[$sKey] = $this->getSafe($sKey);
		return $a;
	}
	
	/// Same as ArrayWrapper::getSafeList, but convert from UTF8.
	public function getSafeListUTF8($sKeys)
	{
		$a = array();
		foreach (getList($sKeys) as $sKey)
			$a[] = utf8::decodeRecurse($this->getSafe($sKey));
		return $a;
	}
	
	
	/// Return an object containing a list of variables
	/// converted to safe format.
	/// @see ArrayWrapper
	public function getSafeObject($sKeys)
	{
		$obj = new stdClass;
		foreach (getList($sKeys) as $sKey)
			$obj->$sKey = &$this->getSafe($sKey);
		return $obj;
	}
	
	
	/// Convert a list of variables to safe format. @see ArrayWrapper
	public function prepare($sKeys, $bUTF8 = false)
	{
		foreach (getList($sKeys) as $sKey)
			$this->prepareOne($sKey, $bUTF8);
	}
	
	/// Convert a variable to safe format. @see ArrayWrapper
	protected function prepareOne($sKey, $bUTF8 = false)
	{
		if (!isset($this->array[$sKey]))
			$this->array[$sKey] = null;
		$value = &$this->array[$sKey];
		forceType($value, $sKey);
		
		if ($bUTF8 && (is_string($value) || substr($sKey, 2) == 'as'))
			$value = utf8::decodeRecurse($value);
	}
	
	
	//-----------------------------------------------------------------
	/// @name Iterator interface: walk the array itself
	/// @{
	//-----------------------------------------------------------------
	
	public function rewind() { reset($this->array); }
	public function current() { return current($this->array); }
	public function key() { return key($this->array); }
	public function next() { return next($this->array); }
	public function valid() { return $this->current() !== false; }
	
	/// @}
	
	
	public function getArray() { return $this->array; }
}

@set_magic_quotes_runtime(false);

/// @cond Hide intermediate variables from Doxygen

$wrapperRequest = new ArrayWrapper($_REQUEST, true);
unset($_REQUEST);
$_REQUEST = &$wrapperRequest;

$wrapperPost = new ArrayWrapper($_POST, true);
unset($_POST);
$_POST = &$wrapperPost;

$wrapperGet = new ArrayWrapper($_GET, true);
unset($_GET);
$_GET = &$wrapperGet;

$wrapperFiles = new ArrayWrapper($_FILES, true);
unset($_FILES);
$_FILES = &$wrapperFiles;

$wrapperServer = new ArrayWrapper($_SERVER, false);
unset($_SERVER);
$_SERVER = &$wrapperServer;

/// @endcond



//-----------------------------------------------------------------
/// Cookies manager
//-----------------------------------------------------------------

class cookies extends ArrayWrapper
{
	public static $iExpireDelaySeconds;
	public static $sPath = '/';
	public static $sDomain = null;
	public static $bHTTPSOnly = false;
	
	protected $iExpireTime = null;
	
	
	public function __construct()
	{
		parent::__construct($_COOKIE, true);
		self::$iExpireDelaySeconds = 24 * 60 * 60;
	}
	
	public function __set($sKey, $value)
	{
		if ($this->iExpireTime === null)
			$this->iExpireTime = time() + self::$iExpireDelaySeconds;
		
		setcookie($sKey, $value, $this->iExpireTime,
			self::$sPath, self::$sDomain, self::$bHTTPSOnly);
	}
	
	public function __unset($sKey)
	{
		if (isset($this->$sKey)) {
			setcookie($sKey, '', time() - 60000);
			$this->$sKey = null;
			$this->prepareOne($sKey);
		}
	}
}

/// @cond Hide intermediate variables from Doxygen

$wrapperCookie = new cookies();
unset($_COOKIE);
$_COOKIE = &$wrapperCookie;

/// @endcond



//-----------------------------------------------------------------
/// @brief Methods to access the member variables of an object
/// in a safe way. Useful when the object comes from JSON.
///
/// The 'prepare' and 'safe' methods convert the variables
/// to a safe format according the following rules:
/// - type: it is set by the forceType() function,
///   according to the variable prefix
/// - default value: if the variable is null,
///   it is given the default value of its type
/// - magic quotes: if the array is concerned by magic quotes,
///   quotes are stripped
///
/// Typical usage:
/// \code
/// $object = new ObjectSafe(...);
/// $object->prepare('sString;bBoolean;iInteger');
/// echo "A string: $object->sString<br />";
/// echo "An integer: $object->iInteger<br />";
/// extract($object->getSafeList('sString;bBoolean;iInteger');
/// echo "A string: $sString<br />";
/// echo "An integer: $iInteger<br />";
/// \endcode
//-----------------------------------------------------------------

class ObjectSafe
{
	/// @brief Create a copy of the given object to ObjectSafe format.
	///
	/// - nested arrays are duplicated
	/// - objects contained in members or arrays are cloned to ObjectSafe
	///
	/// @param $data the object or array to convert.
	public function __construct($data = null)
	{
		if (!is_object($data) && !is_array($data))
			return;
		
		foreach ($data as $sKey => $value) {
			if (is_object($value))
				$this->$sKey = new ObjectSafe($value);
			else{
				if (is_array($value))
					self::convertArray($value);
				$this->$sKey = $value;
			}
		}
	}
	
	/// @brief Convert all elements of the array to ObjectSafe format, recursively:
	/// all objects are cloned to ObjectSafe, recursively.
	private static function convertArray(array &$a)
	{
		foreach ($a as &$value)
			if (is_object($value))
				$value = new ObjectSafe($value);
			else if (is_array($value))
				self::convertArray($value);
	}
	
	
	/// @brief Create a safe object from an integer-indexed array.
	/// The keys of the \$a array are ignored.
	/// The members will take their name from \p $sMembers.
	/// The values will be converted according to their prefix,
	/// with the forceType() function.
	///
	/// @param $a the array to convert.
	/// @param $sMembers the list of the members name. If the list
	/// contains more elements than the \$a array,
	/// the members are set to their empty value.
	public static function fromArray(array $a, $sMembers)
	{
		$obj = new ObjectSafe;
		reset($a);
		foreach (getList($sMembers) as $sMember) {
			list($key, $value) = each($a);
			forceType($value, $sMember);
			$obj->$sMember = $value;
		}
		return $obj;
	}
	
	
	/// @brief Create an object from a safe object.
	/// The members will take their name from \p $sMembers.
	/// The values will be converted according to their prefix,
	/// with the forceType() function.
	///
	/// @param $sClass the name of the object to create.
	/// @param $sMembers the list of the members name.
	/// If \c null, all members are copied to the new object.
	public function toObject($sClass, $sMembers = null)
	{
		$obj = new $sClass;
		
		if ($sMembers === null) {
			foreach (get_object_vars($this) as $sMember => $value) {
				forceType($value, $sMember);
				$obj->$sMember = $value;
			}
		}else{
			foreach (getList($sMembers) as $sMember) {
				list($key, $value) = each($a);
				forceType($value, $sMember);
				$obj->$sMember = $value;
			}
		}
		
		return $obj;
	}
	
	
	/// @brief Return the value of a member in the object.
	/// If the variable is not set, return the given default value.
	public function get($sKey, $defaultValue = null)
	{
		return (isset($this->$sKey))
			? $this->$sKey
			: $defaultValue;
	}
	
	/// Give a default value for a member in the object.
	public function setDefault($sKey, $defaultValue)
	{
		if (!isset($this->$sKey))
			$this->$sKey = $defaultValue;
	}
	
	/// @brief Return the value of a member in the object converted to the given type.
	/// The type name follows the \c settype() function specification.
	public function& getAs($sKey, $sType)
	{
		settype($this->$sKey, $sType);
		return $this->$sKey;
	}
	
	
	/// Convert a member of the object to safe format then return it.
	/// @see ObjectSafe
	public function& getSafe($sKey)
	{
		$this->prepareOne($sKey);
		return $this->$sKey;
	}
	
	/// @brief Return an array containing a list of members
	/// indexed by their name.
	/// @see ObjectSafe
	public function getFields($sKeys)
	{
		$a = array();
		foreach (getList($sKeys) as $sKey)
			$a[$sKey] = $this->getSafe($sKey);
		return $a;
	}
	
	
	/// Convert a list of members to safe format. @see ObjectSafe
	public function prepare($sKeys)
	{
		foreach (getList($sKeys) as $sKey)
			$this->prepareOne($sKey);
	}
	
	/// Convert a member to safe format. @see ObjectSafe
	private function prepareOne($sKey)
	{
		if (!isset($this->$sKey))
			$this->$sKey = null;
		$value = &$this->$sKey;
		forceType($value, $sKey);
	}
}



//-----------------------------------------------------------------
/// Static class for XML writing
//-----------------------------------------------------------------

class xmlwriting
{
	/// @brief Generate an XML tag having the given name, contents and attributes:
	/// <tt>\<name attr="value" ...\>contents\</name\></tt>
	/// @param $sName the name of the tag.
	/// @param $sContents the contents of the tag, should be escaped if not XML code.
	/// @param $aAttr the attributes associative array, formatted by attr::toString().
	/// @param $bNewlineBefore if \c true, insert #sXMLNewline before the tag.
	/// @param $bNewlineAfter if \c true, insert #sXMLNewline after the tag.
	static function tag($sName, $sContents, array $aAttr = array(), $bNewlineBefore = false, $bNewlineAfter = false)
	{
		return
			(($bNewlineBefore) ? sXMLNewline : '') .
			'<' . $sName . attr::toString($aAttr) . '>' . $sContents . '</' . $sName . '>' .
			(($bNewlineAfter)  ? sXMLNewline : '');
	}
	
	/// @brief Generate an XML anchor tag having the given name and attributes:
	/// <tt>\<name attr="value" ... /></tt>
	/// @param $sName the name of the anchor tag.
	/// @param $aAttr the attributes associative array, formatted by attr::toString().
	static function anchor($sName, array $aAttr = array())
	{
		return '<' . $sName . attr::toString($aAttr) . ' />';
	}
	
	
	//-----------------------------------------------------------------
	/// @name Text escaping and quoting
	/// @{
	//-----------------------------------------------------------------
	
	/// HTML-escape a Windows-1252 string.
	static function text($s)
	{
		return htmlspecialchars($s, ENT_QUOTES, sEncoding);
	}
	
	/// HTML-escape the Windows-1252 strings contained in an array.
	static function textArray(array $a)
	{
		return self::map('text', $a);
	}
	
	/// @brief HTML-escape the Windows-1252 strings contained in an array,
	/// but leave the quotes as is.
	static function textNoQuotes($s)
	{
		return htmlspecialchars($s, ENT_NOQUOTES, sEncoding);
	}
	
	/// Apply HTML raw encoding (\c &# prefix) to all characters of a string.
	static function textCoded($s)
	{
		$sCoded = '';
		foreach (str_split($s) as $sChar)
			$sCoded .= '&#' . ord($sChar) . ';';
		return $sCoded;
	}
	
	/// @}
	
	
	/// @brief Map a #xmlwriting function to an array.
	static function map($sFunction, array $a)
	{
		return map(array(__CLASS__, $sFunction), $a);
	}
	
	/// @brief Map a #xmlwriting function to an array, then concatenare the resulting strings.
	static function mapConcat($sFunction, array $a, $sSeparator = '')
	{
		return mapConcat(array(__CLASS__, $sFunction), $a, $sSeparator);
	}
}



//-----------------------------------------------------------------
/// Static class for XML attributes writing
//-----------------------------------------------------------------

class attr
{
	/// @brief Convert an attributes array to a string.
	/// The keys are the attributes name,
	/// the values are the attributes value,
	/// \c null for value-less attributes.
	static function toString(array $aAttr)
	{
		$s = '';
		foreach ($aAttr as $sName => $sValue) {
			$s .= ' ' . $sName;
			if (!is_null($sValue))
				$s .= '="' . html::text($sValue) . '"';
		}
		return $s;
	}
	
	static function className($sClass)  { return ($sClass) ? array('class' => $sClass) : array(); }
	static function id($sId)            { return ($sId)    ? array('id'    => $sId)    : array(); }
	static function style($sStyle)      { return ($sStyle) ? array('style' => $sStyle) : array(); }
	static function name($sName)        { return ($sName)  ? array('name'  => $sName)  : array(); }
	static function idClassName($sId, $sClass) { return self::id($sId) + self::className($sClass); }
	static function idClassNameSame($s) { return self::idClassName($s, $s); }
	static function title($sTitle)      { return ($sTitle) ? array('title' => $sTitle) : array(); }
	
	/// @brief If \p $sName begins with 'ID', create 'name' and 'id' attribute.
	/// If \p $sName does not begin with 'ID', create only 'name' attribute
	/// (no 'id' attribute).
	static function nameid($sName)
	{
		$aAttr = array();
		if (!empty($sName)) {
			if (hasPrefix($sName, 'ID'))
				$aAttr['id'] = $sName = substr($sName, 2);
			$aAttr['name'] = $sName;
		}
		return $aAttr;
	}
}



//-----------------------------------------------------------------
/// Static class for AJAX writing
//-----------------------------------------------------------------

class ajax extends xmlwriting
{
	static function text($s)
	{
		return htmlspecialchars($s, ENT_QUOTES, 'UTF-8');
	}
	
	static function tagtext($sName, $sContents, array $aAttr = array())
	{
		return self::tag($sName, self::text($sContents), $aAttr);
	}
	
	/// @brief Enclose with tag \p $sItemTag each item of the array \p $a.
	/// Return the concatenated XML string.
	static function formatArray($a, $sItemTag)
	{
		$s = '';
		foreach ($a as $item)
			$s .= (is_array($item))
				? self::anchor($sItemTag, $item)
				: self::tag($sItemTag, self::format($item));
		return $s;
	}
	
	
	/// @brief Convert an associative array (tagname => contents, ...) to a list of tags.
	
	/// The array can be recursive.
	/// When the tagname begins with a \c '!', the contents is supposed
	/// to be an XML string.
	/// When the tagname does not begin with a \c '!', the contents is XML-escaped.
	/// If a string is given instead of an array, return it escaped.
	static function format($value)
	{
		if (!is_array($value))
			return self::text($value);
		
		$s = '';
		foreach ($value as $sChildName => $childValue)
			$s .= ($sChildName[0] == '!')
				? self::tag(substr($sChildName, 1), $childValue)
				: self::tag($sChildName, self::format($childValue));
		return $s;
	}

	/// Print an AJAX answer.
	static function output($xml)
	{
		header('Content-type: application/xml');
		echo utf8::encode((is_array($xml)) ? self::format($xml) : $xml);
		exit;
	}
}



//-----------------------------------------------------------------
/// Static class for Perl regexps execution.
//-----------------------------------------------------------------

class regexp
{
	/// @brief Test whether a pattern matches a subject string.
	/// Return the match count, \c false if error.
	public static function test($sPattern, $sSubject, $iFlags = 0, $iOffset = 0)
	{
		$aMatch = array();
		return preg_match($sPattern, $sSubject, $aMatch, $iFlags, $iOffset);
	}
	
	/// @brief Test whether a pattern matches a characters specification,
	/// by using encloseChars() to compute the pattern.
	/// Return the match count, \c false if error.
	public static function testChars($sCharsSpec, $sSubject, $iFlags = 0, $iOffset = 0)
	{
		return self::test(self::encloseChars($sCharsSpec), $sSubject, $iFlags, $iOffset);
	}
	
	
	/// @brief Test whether a pattern matches a subject string.
	/// Return the match count, \c false if error; put the match (if any) in \p $aMatch.
	public static function match($sPattern, $sSubject, array &$aMatch, $iFlags = 0, $iOffset = 0)
	{
		return preg_match($sPattern, $sSubject, $aMatch, $iFlags, $iOffset);
	}
	
	
	/// @brief Get all matches of a subject string against apattern.
	/// Return the match count, \c false if error; put the matches in \p $aMatch.
	public static function matchAll($sPattern, $sSubject, array &$aMatch, $iFlags = 0, $iOffset = 0)
	{
		return preg_match_all($sPattern, $sSubject, $aMatch, $iFlags, $iOffset);
	}
	
	
	/// @brief Replace all occurences of a pattern by a string in a string.
	/// Return the replaced string.
	/// \p $pattern and \p $replacement can be strings or arrays.
	public static function replace($pattern, $replacement, $sSubject, $iLimit = -1)
	{
		return preg_replace($pattern, $replacement, $sSubject, $iLimit);
	}
	
	/// @brief Replace all occurences of a pattern by the result of a callback function.
	/// The callback takes as unique argument the array of matched elements.
	/// Return the replaced string.
	/// \p $pattern and \p $replacement can be strings or arrays.
	public static function replaceCallback($pattern, $callback, $sSubject, $iLimit = -1)
	{
		return preg_replace_callback($pattern, $callback, $sSubject, $iLimit);
	}
	
	/// @brief Replace all occurences of a character specification by a string in a string.
	/// Return the replaced string.
	public static function replaceChars($sCharsSpec, $sReplacement, $sSubject, $iLimit = -1)
	{
		return self::replace(self::encloseChars($sCharsSpec), $sReplacement, $sSubject, $iLimit);
	}
	
	
	/// @brief Split a string against a pattern.
	public static function split($sPattern, $sSubject, $iLimit = -1, $iFlags = 0)
	{
		return preg_split($sPattern, $sSubject, $iLimit, $iFlags);
	}
	
	
	/// @brief Escape special characters in a string to create a valid pattern.
	public static function quote($s)
	{
		return preg_quote($s);
	}
	
	/// @brief Create an enclosed pattern from a pattern.
	public static function enclose($sPattern)
	{
		return '' . $sPattern . '';
	}
	
	/// @brief Create an enclosed pattern from a string to search.
	public static function encloseQuote($s)
	{
		return self::enclose(self::quote($s));
	}
	
	/// @brief Create a single character searching pattern from a character specification.
	public static function encloseChars($sCharsSpec)
	{
		return '[' . $sCharsSpec . ']';
	}
	
	/// @brief Create a single character searching pattern from a characters list.
	public static function encloseQuoteChars($sChars)
	{
		return self::encloseChars(self::quote($sChars));
	}
}



//-----------------------------------------------------------------
/// @brief Basic mutex using files. The PHP \c flock() function
/// is not used, because not working with NFS.
/// Due to FAT time imprecision, the lock can't use FAT files.
/// This class is not 100% secure, but should work in most cases.
//-----------------------------------------------------------------

class file_mutex
{
	private $sFile, $nbTimeoutSeconds;
	private $dtAcquired;
	
	
	/// @brief Initialize the lock but do not acquire it now.
	/// The file should be in a not-FAT and preferably not-NFS file system.
	public function __construct($sFile, $nbTimeoutSeconds)
	{
		$this->sFile            = $sFile;
		$this->nbTimeoutSeconds = $nbTimeoutSeconds;
		$this->dtAcquired       = 0;
	}
	
	/// @brief Release the lock when the object is destructed.
	public function __destruct()
	{
		$this->release();
	}
	
	
	/// @brief Try to acquire the lock. Return \c true if the lock
	/// has been acquired, \c false otherwise.
	public function tryAcquire()
	{
		clearstatcache();
		
		$dtNow = time();
		$dtLimit = $dtNow - nbMutexTimeoutSeconds;
		
		if ($this->dtAcquired >= $dtLimit) {
			// Lock already acquired: refresh it
			
			$this->dtAcquired = $dtNow;
			touch($this->sFile, $dtNow);
			return true;
		}
		
		// Get the lock file modification time.
		// If the file exists and is recent, consider the lock as acquired.
		$dtFile = @filemtime($this->sFile);
		if ($dtFile !== false && $dtFile >= $dtLimit)
			return false;
		
		// We suppose the touch() function is atomic.
		if (!touch($this->sFile, $dtNow)) {
			// Maybe not acquired: need to test
			sleep(1);
			if (@filemtime($this->sFile) != $dtNow)
				return false;
		}
		
		// Lock acquired
		$this->dtAcquired = $dtNow;
		return true;
	}
	
	
	/// @brief Release the lock if it is acquired.
	public function release()
	{
		clearstatcache();
		
		if (@filemtime($this->sFile) == $this->dtAcquired)
			@unlink($this->sFile);
	}
}



/// @brief Set the output MIME type to text.
function useMimeText()
{
	header('Content-type: text/plain');
}


?>