<?php defined('KERNEL_INCLUDED') or die('kernel should be included first');
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder

//-----------------------------------------------------------------
/// @file
/// @brief #Page: XHTML page base class
//-----------------------------------------------------------------


//-----------------------------------------------------------------
/// XHTML page base class
//-----------------------------------------------------------------

abstract class Page
{
	/// @brief Text that will be replaced by the page last modification time.
	/// If the page is cached, it won't be updated until the page is generated again.
	const sLastModificationTimeTag = '[[[LAST_MODIFICATION_TIME]]]';
	
	/// @brief Format given to date() function to print the page last modification time.
	const sLastModificationTimeFormat = 'Y-m-d';
	
	/// @brief Text that will be replaced by the page generation duration.
	/// If the page is cached, it won't be updated until the page is generated again.
	const sGenerationTimeTagInitial = '[[[GENERATION_TIME_INITIAL]]]';
	
	/// @brief Text that will be replaced by the page printing.
	/// It is updated each time the page is printed.
	const sGenerationTimeTag = '[[[GENERATION_TIME]]]';
	
	
	private static $currentPage = null;
	
	/// Reference to the Page object being printed.
	public static function currentPage() { return $currentPage; }
	
	
	private $aCacheDependency = array();
	private $canCache = true;
	
	
	/// @brief Called by the Page::all() method, just before page code computing.
	///
	/// Is called even if the page is read from cache and does not need to be rebuild.
	/// Thus, this method can be used to enable/disable caching dynamically.
	protected function initialize() {}
	
	/// Title written in <tt>\<title\></tt> tag.
	protected abstract function getTitle();
	
	/// CSS file to load.
	protected function getCSS() { return null; }
	
	/// Inline CSS code to write.
	protected function getInlineCSS() { return null; }
	
	
	//-----------------------------------------------------------------
	/// @name Meta fields
	/// @{
	//-----------------------------------------------------------------
	
	protected function getIcon()        { return null; }
	protected function getAuthor()      { return null; }
	protected function getDescription() { return null; }
	protected function getKeywords()    { return null; }
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Caching
	/// Overload Page::isCachingEnabled() to enable caching.
	/// Call Page::addCurrentCacheDependency() to notify the kernel
	/// about a dependency.
	/// @{
	//-----------------------------------------------------------------
	
	protected function isCachingEnabled() { return false; }
	protected function canCache() { return $this->isCachingEnabled() && $this->canCache; }
	protected function getCacheFile() { return basename($_SERVER->SCRIPT_FILENAME . '.cache'); }
	
	public function addCacheDependency($sFileName)
	{
		if (!in_array($sFileName, $this->aCacheDependency))
			$this->aCacheDependency[] = $sFileName;
	}
	
	public static function addCurrentCacheDependency($sFileName, $bSetCanCache)
	{
		if (self::$currentPage) {
			self::$currentPage->addCacheDependency($sFileName);
			self::$currentPage->canCache = true;
		}
	}
	
	protected function needCacheRebuild()
	{
		$sCacheFile = $this->getCacheFile();
		if (!is_file($sCacheFile)) {
			if (bDebugCache)
				echo html::div('Cache not found: ' . $sCacheFile);
			return true;
		}
		
		$dtCache = filemtime($sCacheFile);
		foreach ($this->aCacheDependency as $sFile) {
			if ($dtCache < filemtime($sFile)) {
				if (bDebugCache)
					echo html::div('Cache fault: cache time: ' . date('r', $dtCache) . ' - ' . $sFile . ' time ' . date('r', filemtime($sFile)));
				return true;
			}
		}
		return false;
	}
	
	/// @}
	
	
	//-----------------------------------------------------------------
	/// @name Contents
	/// @{
	//-----------------------------------------------------------------
	
	/// Return the CSS related code to be put in the <tt>\<head\></tt> tag.\
	/// Concatenate linked CSS (Page::getCSS) and inline CSS (Page::getInlineCSS).
	protected function css()
	{
		$sLinkedCSS = $this->getCSS();
		if ($sLinkedCSS)
			$sLinkedCSS = html::cssLink($sLinkedCSS);
		
		$sInlineCSS = $this->getInlineCSS();
		if ($sInlineCSS)
			$sInlineCSS = html::cssInline($sInlineCSS);
		
		return $sLinkedCSS . $sInlineCSS;
	}
	
	
	/// @brief Content to be put in the <tt>\<head\></tt> tag.
	/// Handle <tt>\<title\></tt>, various <tt>\<meta\></tt>, and CSS.
	protected function head()
	{
		return
			html::tag('title', $this->getTitle()) .
			html::metaHTTP('Content-Type', 'text/html; charset=' . sEncoding) .
			html::icon($this->getIcon()) .
			html::meta('Author',      $this->getAuthor()) .
			html::meta('Description', $this->getDescription()) .
			html::meta('Keywords',    $this->getKeywords()) .
			$this->css();
	}
	
	
	/// Return the scripts HTML code, written inside the <tt>\<body\></tt> tag.
	protected function scripts() { return ''; }
	
	
	/// Return the page contents, in a <tt>\<body\></tt> tag.
	protected abstract function body();
	
	
	/// @brief Return the whole page code (from cache if possible), including DTD.
	///
	/// - Determine whether the page can be read from cache
	/// - If possible, verify it is not outdated
	/// - If generation is required, do it
	/// - If caching is enabled, put the caching duration
	///   (#sGenerationTimeTagInitial) and save the result to the cache
	/// - Put the printing duration (#sGenerationTimeTag)
	public function all()
	{
		self::$currentPage = $this;
		
		try{
			$this->initialize();
		}catch(PageException $e) {
		}
		append($this->aCacheDependency, get_included_files());
		
		
		$fromCache = $this->canCache() && !$this->needCacheRebuild();
		if ($fromCache) {
			$cache = unserialize(@file_get_contents($this->getCacheFile()));
			$s = $cache->sText;
			$this->aCacheDependency = $cache->aDependency;
			$fromCache = !$this->needCacheRebuild();
			if (bDebugCache && $fromCache)
				echo 'From cache' . html::br();
		}
		
		if (!$fromCache) {
			$this->canCache = false;
			$s = $this->html();
			$s = replace($s, array(
				self::sLastModificationTimeTag      => @date(self::sLastModificationTimeFormat, getlastmod()),
				self::sGenerationTimeTagInitial => elapsedTime()));
		}
		
		append($this->aCacheDependency, get_included_files());
		makeUnique($this->aCacheDependency);
		
		if (bDebugCache)
			echo debug::dump($this->aCacheDependency), debug::dump($this->canCache()), debug::dump($fromCache);
		
		if (!$fromCache && $this->canCache()) {
			if (bDebugCache)
				echo 'Caching now!' . html::br();
			$cache = new stdClass;
			$cache->sText = $s;
			$cache->aDependency = $this->aCacheDependency;
			file_put_contents($this->getCacheFile(), serialize($cache));
		}
		
		$s = str_replace(self::sGenerationTimeTag, elapsedTime(), $s);
		
		self::$currentPage = null;
		return $s;
	}
	
	
	/// Generate the whole page HTML code, according to Page::isFrameset().
	///
	/// The functions are not called in the HTML order:
	/// - body
	/// - head
	/// - scripts
	///
	/// This allows the page to make its title or its header
	/// depend of the page contents.
	protected function html()
	{
		$sBody   = $this->body();
		$sHead   = $this->head();
		$sScript = $this->scripts();
		return ($this->isFrameset())
			? htmlframeset::sDocType .
				htmlframeset::tag('html',
					htmlframeset::tag('head', htmlframeset::convertAnchors($sHead . $sScript)) .
					htmlframeset::convertAnchors($sBody))
			: sDocType .
				html::tag('html',
					html::tag('head', $sHead, array(), true, true) .
					html::tag('body', $sScript . $sBody),
					array(), true, false);
	}
	
	/// @brief Indicate if the page is a frameset.
	/// Used by #html() to use the right DTD.
	protected function isFrameset() { return false; }
	
	/// @}
	
	
	/// Print a page, with a <tt>Content-type</tt> HTTP header.
	public static function output(Page $page)
	{
		$s = $page->all();
		
		self::mimeType('text/html; charset=' . sEncoding);
		echo $s;
	}
	
	
	//-----------------------------------------------------------------
	/// @name Helpers
	/// @{
	//-----------------------------------------------------------------
	
	/// Use default arguments to use online W3C validator.
	protected function getValidator($sValidatorURL = 'validator.w3.org', $sImageURL = 'www.w3.org')
	{
		return html::link('http://' . $sValidatorURL . '/check?uri=referer',
			html::img('http://' . $sImageURL . '/Icons/valid-xhtml10', 'Valid XHTML 1.0!', 88,31));
	}
	
	
	protected function getGoogleAnalytics($sID)
	{
		return <<<EOF
<script src="http://www.google-analytics.com/urchin.js" type="text/javascript"></script>
<script type="text/javascript">_uacct = "$sID"; urchinTracker();</script>
EOF;
	}
	
	/// @brief Return Google Search hidden form fields and scripts.
	/// @param $aFormAttr array where to put the form attributes.
	/// @param $sGoogleSearchId Google Co-op ID.
	/// @param $iFORID used to compute form height. Valid values are: 9, 10, 11. Use 10 or 11 if displaying ads.
	protected function getGoogleCSEFormHidden(array& $aFormAttr, $sGoogleSearchId, $iFORID = 9)
	{
		$aFormAttr += attr::id('searchbox_' . $sGoogleSearchId);
		return
			form::hidden('cx', $sGoogleSearchId) .
			form::hidden('cof', 'FORID:' . $iFORID) .
			html::jscriptLink('http://www.google.com/coop/cse/brand?form=searchbox_' . urlencode($sGoogleSearchId));
	}
	
	/// Print Google Search results.
	protected function getGoogleCSEResults($sGoogleSearchId, $aJscriptVar = array())
	{
		$sidResults = 'results_' . $sGoogleSearchId;
		append($aJscriptVar, array(
			'googleSearchIframeName' => $sidResults,
			'googleSearchFormName'   => 'searchbox_' . $sGoogleSearchId,
			'googleSearchDomain'     => 'www.google.com',
			'googleSearchPath'       => '/cse',
		));
		
		return
			html::div('', attr::id($sidResults)) .
			html::jscriptInline(html::jscriptSetVars($aJscriptVar)) .
			html::jscriptLink('http://www.google.com/afsonline/show_afs_search.js');
	}
	
	/// Prevent web browsers and proxies from caching the page.
	public static function preventCaching()
	{
		header('Expires: 0');
		header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT');
		header('Cache-Control: no-store, no-cache, must-revalidate');
		header('Pragma: no-cache');
	}
	
	/// Specify a MIME type for the page
	public static function mimeType($sMimeType)
	{
		if ($sMimeType)
			header('Content-type: ' . $sMimeType);
	}
	
	
	/// Prepare a file upload to the client.
	///
	/// @param $sFileName the file name as it will appear to the client,
	/// \c null to try to display the file in the browser instead of downloading it.
	/// @param $sMimeType force the MIME type of the file, \c null to let the browser guess.
	public static function uploadData($sFileName, $sMimeType = null)
	{
		if ($sMimeType)
			self::mimeType($sMimeType);
		
		/// Disable compression for IE
		if (strpos($_SERVER->get('HTTP_USER_AGENT'), '/MSIE/') !== false)
			ini_set('zlib.output_compression', 'Off');
		
		if ($sFileName !== null)
			header('Content-Disposition: attachment; filename="' . $sFileName . '"');
		
		self::preventCaching();
	}
	
	
	/// @brief Upload a file to the client.
	/// Return \c true on success, \c false on failure.
	///
	/// @param $sPathName the path of the file to upload.
	/// @param $sMimeType the MIME type of the file to upload, \c null to detect it.
	/// @param $sFileName the file name as it will appear to the client,
	/// \c null to deduce it from \p $sPathName.
	/// @param $bForceFile if \c true, force the file to be downloaded
	/// by the download manager; if \c true, browser-supported files
	/// (HTML, text, images) will be displayed instead of downloaded.
	public static function uploadFile($sPathName, $sMimeType = null, $sFileName = null, $bForceFile = true)
	{
		if (!@is_readable($sPathName))
			return false;
		
		if ($sFileName === null)
			$sFileName = basename($sPathName);
		self::uploadData(($bForceFile) ? $sFileName : null, $sMimeType);
		
		// Disable caching
		header('Cache-control: private');
		header('Pragma: private');
		
		// Put the file size
		$size = @filesize($sPathName);
		if ($size !== false)
			header('Content-Length: ' . $size);
		
		$file = @fopen($sPathName, 'r');
		if (!$file)
			return false;
		
		set_time_limit(0);
		while (!feof($file) && !connection_status()) {
			echo fread($file, 8 << 10);
			flush();
		}
		fclose($file);
		return !connection_status();
	}
	
	
	public static function redirect($sURL)
	{
		header('Location: ' . $sURL);
		exit;
	}
	
	///@}
	
	
	//-----------------------------------------------------------------
	/// @name Parameters
	/// @{
	//-----------------------------------------------------------------
	
	/// @brief Load values from \c $_REQUEST and store them as member in the object.
	/// The parameters should be \c public properties of the class.
	protected function loadRequestParams($sList)
	{
		foreach (getList($sList) as $sName)
			$this->$sName = $_REQUEST->getSafe($sName);
	}
	
	///@}
}


/// Exception caught by #Page class when executing the page::initialize() method
class PageException extends Exception
{
}

?>