<?php
// This file is part of DENIF, distributed under license GPLv3.
// Copyright 2006 Guillaume Ryder, Arthur Charguraud

//-----------------------------------------------------------------
/// @file
/// @brief Download documents and save them in the cache
///
/// Files are downloaded only for documents matching all requirements:
/// - node::$iPingDownload is not \c pdlIgnore or node::$dtPing is unspecified
/// - URL specified
/// - directories created,
/// - all node::$sPathPart of the ancestors specified
//-----------------------------------------------------------------


require_once 'shared.inc.php';


//-----------------------------------------------------------------
/// @brief Base class for ping/download scripts.
/// The PingDownload::progressPrint() method must be overloaded
/// when printing HTML code.
//-----------------------------------------------------------------

class PingDownload
{
	/// Default verbosity limit for printed messages.
	const iVerbosity = 1;
	
	
	/// @brief Execute the ping/download operation.
	/// Print the progression messages as plain text, then quit.
	public static function executePlainText()
	{
		useMimeText();
		$pingDownload = new PingDownload();
		$pingDownload->processExclusive();
		exit(0);
	}
	
	
	/// @brief Print a progress message, or not, according to its verbosity level.
	/// By default, print the message as is (for command line).
	protected function progressPrint($iVerbosity, $sLine)
	{
		if ($iVerbosity <= self::iVerbosity)
			echo $sLine;
	}
	
	/// @brief Flush the progression messages.
	protected function progressFlush()
	{
		flush();
	}
	
	
	/// @brief Ping/download the files, only if not other ping/download process is running.
	public function processExclusive()
	{
		$mutex = new file_mutex(sMutexFile, nbMutexTimeoutSeconds);
		if ($mutex->tryAcquire()) {
			$this->process();
			$mutex->release();
		}
	}
	
	
	/// @brief Format a file size into a string, in megabytes.
	private static function formatSize($iSizeBytes)
	{
		return sprintf('%5.2f Mb', $iSizeBytes / (1 << 20));
	}
	
	
	/// @brief Ping/download the files.
	public function process()
	{
		$sTable = node::$table->sName;
		$dtLimit = time() + maxTimeScriptsSeconds;
		set_time_limit(maxTimeScriptsSecondsHard);
		
		$aURLTested = array();
		$nbFailure      = 0;
		$nbProcessed    = 0;
		$lenCopiedTotal = 0;
		
		do{
			
			sql::lockTables(array($sTable));
			
			$node = sql::selectOneFetch($sTable, array('*'),
				'WHERE iNode = ' . nodeLast . ' AND sURL <> "" AND sPathPart <> "" AND (iPingDownload <> ' . pdlIgnore . ' OR dtPing = 0) LIMIT 1',
				$sTable);
			
			if ($node)
				sql::update($sTable, array('dtPing' => time(), 'iPingDownload' => pdlIgnore), sql::whereId($node->id));
			
			sql::unlockTables();
			
			if (!$node)
				break;
			
			$target = target::fromURL($node->sURL);
			$target->removeParamsFromPrefix(sParamPrefix);
			$sURL = $target->url();
			
			$bOK = true;
			foreach ($node->loadAllParents(true) as $nodeParent)
				if (!$nodeParent->sPathPart) {
					$bOK = false;
					break;
				}
			$sLocalFile = ($bOK) ? $node->sPathFull : '';
			
			clearstatcache();
			
			$bPingOK = defaultElement($aURLTested, $sURL, true);
			$bCanDownload = ($node->iPingDownload == pdlDownload) && $sLocalFile;
			
			$this->progressPrint(1, (($bCanDownload) ? 'Downloading' : 'Pinging') . ' document - ID: ' . $node->id . ', URL: ' . $sURL);
			$this->progressPrint(2, CRLF . 'To: ' . $sLocalFile . CRLF);
			$this->progressFlush();
			
			$aUpdate = array();
			
			$bOK     = false;
			$sStatus = 'failure: URL not found';
			if ($bPingOK) {
				$fileInput = @fopen($sURL, 'rb');
				$bPingOK = ($fileInput != null);
				if ($bPingOK && $bCanDownload) {
					
					// Get a non-existing file name to store the downloaded file to
					$sLocalDir = path::dirname($sLocalFile);
					@path::mkdir($sLocalDir, iDataUnixRightsDirs, true);
					$sLocalFileTemp = path::getTempFileName(sTempFilePrefix, $sLocalDir);
					$fileOutput = @fopen($sLocalFileTemp, 'wb');
					if ($fileOutput) {
						$this->progressPrint(2, 'Downloading to: ' . $sLocalFileTemp . CRLF);
						
						@stream_copy_to_stream($fileInput, $fileOutput, maxCacheFileSize + 1);
						fclose($fileOutput);
						$lenCopied = @filesize($sLocalFileTemp);
						
						$this->progressPrint(2, 'Downloaded to: ' . $sLocalFile . CRLF);
						$this->progressPrint(2, 'Temporary file: ' . $sLocalFileTemp . CRLF);
						
						if ($lenCopied > maxCacheFileSize) {
							// File too big: forget it
							
							$sStatus = 'failure: file too big';
							unlink($sLocalFileTemp);
							
						}else{
							// File OK: overwrite the cached file
							
							$bOK = true;
							$sStatus = 'success: ' . self::formatSize($lenCopied);
							$this->progressPrint(2, '  Renaming: ' . $sLocalFileTemp . CRLF . '  To: ' . $sLocalFile . CRLF);
							
							@unlink($sLocalFile);
							@rename($sLocalFileTemp, $sLocalFile);
							@chmod($sLocalFile, iDataUnixRightsFiles);
							@chgrp($sLocalFile, sDataUnixGroup);
							$lenCopiedTotal += $lenCopied;
							
							$aUpdate['bCached'] = true;
						}
					}
					
					fclose($fileInput);
				}
			}
			
			$aURLTested[$sURL] = $bPingOK;
			
			$this->progressPrint(1, '  ' . $sStatus . CRLF);
			$this->progressPrint(2, CRLF);
			
			$aUpdate['bPingOK'] = $bPingOK;
			$nbProcessed++;
			sql::update($sTable, $aUpdate, sql::whereId($node->id));
			
		}while (time() < $dtLimit);
		
		$this->progressPrint(0,
			'Done.' . CRLF .
			'Files processed: ' . sprintf('%4d', $nbProcessed) . CRLF .
			'Failures:        ' . sprintf('%4d', $nbFailure) . CRLF .
			'Copied:          ' . self::formatSize($lenCopiedTotal));
		$this->progressFlush();
	}
}


// Entry point
if (bCmdLine)
	PingDownload::executePlainText();

?>